# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import platform

import pkg_resources

from swagger_spec_compatibility.cli.common import add_rule_discovery_argument
from swagger_spec_compatibility.cli.common import cli_rules
from swagger_spec_compatibility.cli.common import CLIProtocol
from swagger_spec_compatibility.rules import RuleRegistry


def execute(cli_args):  # pragma: no cover
    # type: (CLIProtocol) -> int
    print(
        'swagger-spec-compatibility: {swagger_spec_compatibility_version}\n'
        'Python version: {python_implementation} - {python_version}\n'
        'Python compiler: {python_compiler}\n'
        'Discovered rules:\n    {rules}'.format(
            swagger_spec_compatibility_version=pkg_resources.get_distribution('swagger-spec-compatibility').version,
            python_implementation=platform.python_implementation(),
            python_version=platform.python_version(),
            python_compiler=platform.python_compiler(),
            rules='\n    '.join(
                '{rule_name}: {rule.__module__}.{rule.__name__}'.format(rule_name=rule_name, rule=RuleRegistry.rule(rule_name))
                for rule_name in cli_rules()
            ),
        ),
    )

    return 0


def add_sub_parser(subparsers):
    # type: (argparse._SubParsersAction) -> argparse.ArgumentParser
    info_parser = subparsers.add_parser('info', help='Reports tool\'s information')
    add_rule_discovery_argument(info_parser)
    return info_parser



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing

from bravado_core.spec import Spec

from swagger_spec_compatibility.rules.common import BaseRule
from swagger_spec_compatibility.rules.common import Level
from swagger_spec_compatibility.rules.common import RuleType
from swagger_spec_compatibility.rules.common import ValidationMessage
from swagger_spec_compatibility.util import is_path_in_top_level_paths
from swagger_spec_compatibility.walkers import format_path
from swagger_spec_compatibility.walkers.additional_properties import AdditionalPropertiesDifferWalker
from swagger_spec_compatibility.walkers.additional_properties import DiffType
from swagger_spec_compatibility.walkers.request_parameters import RequestParametersWalker


class RemovedPropertiesFromRequestObjectsWithAdditionalPropertiesSetToFalse(BaseRule):
    description = \
        'If the object is defined with additionalProperties set to False then the object will not allow presence of ' \
        'properties not defined on the properties section of the object definition. Removing a definition of an ' \
        'existing property makes objects sent from a client, that is using "old" Swagger specs, to the server be ' \
        'considered invalid by the backend.'
    error_code = 'REQ-E003'
    error_level = Level.ERROR
    rule_type = RuleType.REQUEST_CONTRACT
    short_name = 'Removing properties from an object with additionalProperties set to False used as request parameter'

    @classmethod
    def validate(cls, left_spec, right_spec):
        # type: (Spec, Spec) -> typing.Iterable[ValidationMessage]
        request_parameter_paths = RequestParametersWalker(left_spec, right_spec).walk()
        for additional_properties_diff in AdditionalPropertiesDifferWalker(left_spec, right_spec).walk():
            if additional_properties_diff.diff_type != DiffType.PROPERTIES:
                continue
            if additional_properties_diff.properties and not additional_properties_diff.properties.old:
                continue
            if not is_path_in_top_level_paths(request_parameter_paths, additional_properties_diff.path):
                continue
            yield cls.validation_message(format_path(additional_properties_diff.path))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing

from bravado_core.spec import Spec

from swagger_spec_compatibility.rules.common import RuleProtocol
from swagger_spec_compatibility.rules.common import RuleRegistry
from swagger_spec_compatibility.rules.common import ValidationMessage


class _ALL_RULES(object):
    def __str__(self):
        # type: () -> str
        return str('ALL_RULES')  # pragma: no cover  # This statement is present only to have a nicer REPL experience


def compatibility_status(
    old_spec,  # type: Spec
    new_spec,  # type: Spec
    rules=_ALL_RULES(),  # type: typing.Union[_ALL_RULES, typing.Iterable[typing.Type[RuleProtocol]]]
):
    # type: (...) -> typing.Mapping[typing.Type[RuleProtocol], typing.Iterable[ValidationMessage]]

    if isinstance(rules, _ALL_RULES):
        rules = RuleRegistry.rules()

    rules_to_error_level_mapping = {
        rule: list(rule.validate(left_spec=old_spec, right_spec=new_spec))
        for rule in rules
    }

    return rules_to_error_level_mapping



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import os
import typing
import warnings
from argparse import ArgumentTypeError
from importlib import import_module
from os.path import abspath
from os.path import exists
from os.path import expanduser
from os.path import expandvars

import typing_extensions
from six.moves.urllib.parse import urljoin
from six.moves.urllib.parse import urlsplit
from six.moves.urllib.request import pathname2url
from six.moves.urllib.request import url2pathname
from venusian import Scanner

from swagger_spec_compatibility.rules.common import BaseRule
from swagger_spec_compatibility.rules.common import RuleRegistry


_ENV_VARIABLE = str('CUSTOM_RULE_PACKAGES')


class CLIProtocol(typing_extensions.Protocol):
    command = None  # type: typing.Text
    func = None  # type: typing.Callable[['CLIProtocol'], int]


class CLIRulesProtocol(typing_extensions.Protocol):
    rules = None  # type: typing.Iterable[typing.Text]
    blacklist_rules = None  # type: typing.Iterable[typing.Text]


def uri(param):
    # type: (str) -> str
    if urlsplit(param).scheme:
        return param

    path = expanduser(expandvars(url2pathname(param)))
    if exists(path):
        return urljoin('file:', pathname2url(abspath(path)))

    raise ArgumentTypeError('`{param}` is not an existing file and either a valid URI'.format(param=param))


def cli_rules():
    # type: () -> typing.List[typing.Text]
    return list(RuleRegistry.rule_names())


def rules(cli_args):
    # type: (CLIRulesProtocol) -> typing.Set[typing.Type[BaseRule]]
    return {
        RuleRegistry.rule(rule_name)
        for rule_name in cli_args.rules
        # The parser defines rules and blacklist_rules as mutual exclusive
        # so we're guaranteed to have at least one set to it's default value
        if rule_name not in cli_args.blacklist_rules
    }


def _get_rule_discovery_from_env():
    # type: () -> typing.List[typing.Text]
    return [
        value
        for value in (
            part.strip()
            for part in os.environ.get(_ENV_VARIABLE, '').split(',')
        )
        if value
    ]


def add_rule_discovery_argument(argument_parser):
    # type: (argparse.ArgumentParser) -> None
    argument_parser.add_argument(
        '-d', '--discover-rules-from',
        action='append',
        default=[_get_rule_discovery_from_env()],
        dest='packages_to_discover_rules_from',
        help='Non-standard packages to load rules from. '
             'The values can be defined as CSV in the {} '
             'environmental variable'.format(_ENV_VARIABLE),
        nargs='+',
    )


def pre_process_cli_to_discover_rules(argv=None):
    # type: (typing.Optional[typing.Sequence[typing.Text]]) -> None
    scanner = Scanner()
    rule_discovery_parser = argparse.ArgumentParser(add_help=False)
    add_rule_discovery_argument(rule_discovery_parser)
    rule_discovery_args, _ = rule_discovery_parser.parse_known_args(argv)
    for packages in (rule_discovery_args.packages_to_discover_rules_from or []):
        for package in packages:
            try:
                scanner.scan(import_module(package))
            except ImportError:
                warnings.warn(
                    RuntimeWarning('\'{}\' module, specified via the rule discovery CLI, is not found. Ignoring it.'.format(package)),
                )


def add_rules_arguments(argument_parser):
    # type: (argparse.ArgumentParser) -> None
    rules = cli_rules()
    if not rules:
        raise argument_parser.error('No rules are defined.')

    add_rule_discovery_argument(argument_parser)

    mutex_group = argument_parser.add_mutually_exclusive_group()
    mutex_group.add_argument(
        '-r', '--rules',
        action='append',
        choices=rules,
        dest='rules',
        help='Rules to apply for compatibility detection. (default: [%(choices)s])',
        metavar='RULE',
        nargs='+',
    )
    mutex_group.add_argument(
        '-b', '--blacklist-rules',
        action='append',
        choices=rules,
        dest='blacklist_rules',
        help='Rules to ignore for compatibility detection. (By default no rules are blacklisted)',
        metavar='RULE',
        nargs='+',
    )


def post_process_rules_cli_arguments(args):
    # type: (argparse.Namespace) -> argparse.Namespace
    def _extract_rules(field, default_value):
        # type: (typing.Text, typing.Iterable[typing.Text]) -> None
        if hasattr(args, field):  # pragma: no branch
            if getattr(args, field):
                setattr(
                    args, field,
                    list({rule for rules in getattr(args, field) for rule in rules}),
                )
            else:
                setattr(args, field, default_value)

    _extract_rules('rules', cli_rules())
    _extract_rules('blacklist_rules', [])

    return args



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing

from bravado_core.spec import Spec

from swagger_spec_compatibility.rules.common import BaseRule
from swagger_spec_compatibility.rules.common import Level
from swagger_spec_compatibility.rules.common import RuleType
from swagger_spec_compatibility.rules.common import ValidationMessage
from swagger_spec_compatibility.util import is_path_in_top_level_paths
from swagger_spec_compatibility.walkers import format_path
from swagger_spec_compatibility.walkers.required_properties import RequiredPropertiesDifferWalker
from swagger_spec_compatibility.walkers.response_paths import ResponsePathsWalker


class RemovedRequiredPropertyFromResponse(BaseRule):
    description = \
        'Removing a required property from an object leads to false expectation on the client receiving the object. ' \
        'If the client is using "old" service\'s Swagger spec it will expect the property to be present and so it ' \
        'could throw errors. It could be valid to assume that the client won\'t perform response validation and this ' \
        'to unexpected errors while parsing the response and/or using the missing property.'
    error_level = Level.ERROR
    error_code = 'RES-E002'
    rule_type = RuleType.RESPONSE_CONTRACT
    short_name = 'Removed Required Property from Response contract'

    @classmethod
    def validate(cls, left_spec, right_spec):
        # type: (Spec, Spec) -> typing.Iterable[ValidationMessage]
        response_paths = ResponsePathsWalker(left_spec, right_spec).walk()
        for required_property_diff in RequiredPropertiesDifferWalker(left_spec, right_spec).walk():
            if not required_property_diff.mapping.old:
                continue
            if not is_path_in_top_level_paths(response_paths, required_property_diff.path):
                continue
            yield cls.validation_message(format_path(required_property_diff.path))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing

from bravado_core.spec import Spec

from swagger_spec_compatibility.rules.common import BaseRule
from swagger_spec_compatibility.rules.common import Level
from swagger_spec_compatibility.rules.common import RuleType
from swagger_spec_compatibility.rules.common import ValidationMessage
from swagger_spec_compatibility.spec_utils import get_endpoints


class DeletedEndpoint(BaseRule):
    description = \
        'An endpoint has been removed. This change is not backward compatible as holders of stale swagger ' \
        'specs (like old mobile Apps) could continue to call the removed endpoint and this will cause an ' \
        'HTTP error status code (usually an HTTP/400 or HTTP/404)'
    error_code = 'MIS-E001'
    error_level = Level.ERROR
    rule_type = RuleType.MISCELLANEOUS
    short_name = 'Delete Endpoint'

    @classmethod
    def validate(cls, left_spec, right_spec):
        # type: (Spec, Spec) -> typing.Iterable[ValidationMessage]
        endpoints_left_spec = get_endpoints(left_spec)
        endponts_right_spec = get_endpoints(right_spec)

        return (
            cls.validation_message(str(removed_endpoint))
            for removed_endpoint in endpoints_left_spec - endponts_right_spec
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing
from itertools import chain

from bravado_core.spec import Spec
from bravado_core.util import determine_object_type
from bravado_core.util import ObjectType

from swagger_spec_compatibility.walkers import NoValue
from swagger_spec_compatibility.walkers import PathType
from swagger_spec_compatibility.walkers import SchemaWalker


class RequestParametersWalker(SchemaWalker[PathType]):
    # TODO: update the name as it gets only the schemas of the parameters
    left_spec = None  # type: Spec
    right_spec = None  # type: Spec

    def fix_parameter_path(self, path, original_path, value):
        # type: (PathType, PathType, PathType) -> PathType
        return tuple(chain(original_path, value[len(original_path):]))

    def should_path_be_walked_through(self, path):
        # type: (PathType) -> bool
        if not path:
            return True

        # Request parameters could be defined in
        # - ('paths', endpoint, 'parameters', idx, 'schema')
        # - ('paths', endpoint, http_verb, 'parameters', idx, 'schema')
        if path[0] != 'paths':
            return False

        if len(path) >= 4 and (path[2] != 'parameters' and path[3] != 'parameters'):
            return False

        if len(path) >= 6 and (path[4] != 'schema' and path[5] != 'schema'):
            return False

        return True

    def __init__(
        self,
        left_spec,  # type: Spec
        right_spec,  # type: Spec
    ):
        # type: (...) -> None
        super(RequestParametersWalker, self).__init__(left_spec=left_spec, right_spec=right_spec)

    def dict_check(
        self,
        path,  # type: PathType
        left_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
        right_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
    ):
        # type: (...) -> typing.Iterable[PathType]
        if determine_object_type(left_dict) == ObjectType.PARAMETER or determine_object_type(right_dict) == ObjectType.PARAMETER:
            return (path, )
        else:
            return ()

    def list_check(
        self,
        path,  # type: PathType
        left_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
        right_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
    ):
        # type: (...) -> typing.Iterable[PathType]
        return ()

    def value_check(
        self,
        path,  # type: PathType
        left_value,  # type: typing.Any
        right_value,  # type: typing.Any
    ):
        # type: (...) -> typing.Iterable[PathType]
        return ()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing

from bravado_core.spec import Spec

from swagger_spec_compatibility.rules.common import BaseRule
from swagger_spec_compatibility.rules.common import Level
from swagger_spec_compatibility.rules.common import RuleType
from swagger_spec_compatibility.rules.common import ValidationMessage
from swagger_spec_compatibility.util import is_path_in_top_level_paths
from swagger_spec_compatibility.walkers import format_path
from swagger_spec_compatibility.walkers.additional_properties import AdditionalPropertiesDifferWalker
from swagger_spec_compatibility.walkers.additional_properties import DiffType
from swagger_spec_compatibility.walkers.response_paths import ResponsePathsWalker


class AddedPropertiesInResponseObjectsWithAdditionalPropertiesSetToFalse(BaseRule):
    description = \
        'If the object is defined with additionalProperties set to False then the object will not allow presence of ' \
        'properties not defined on the properties section of the object definition. Adding a definition of a new ' \
        'property makes object sent from the server to the client be considered invalid by a client that is using ' \
        '"old" Swagger specs.'
    error_code = 'RES-E001'
    error_level = Level.ERROR
    rule_type = RuleType.RESPONSE_CONTRACT
    short_name = 'Added properties in an object with additionalProperties set to False used in response'

    @classmethod
    def validate(cls, left_spec, right_spec):
        # type: (Spec, Spec) -> typing.Iterable[ValidationMessage]
        response_paths = ResponsePathsWalker(left_spec, right_spec).walk()
        for additional_properties_diff in AdditionalPropertiesDifferWalker(left_spec, right_spec).walk():
            if additional_properties_diff.diff_type != DiffType.PROPERTIES:
                continue
            if additional_properties_diff.properties and not additional_properties_diff.properties.new:
                continue
            if not is_path_in_top_level_paths(response_paths, additional_properties_diff.path):
                continue
            yield cls.validation_message(format_path(additional_properties_diff.path))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing

try:
    from functools import lru_cache as _lru_cache  # type: ignore # py>=3.2
except ImportError:  # pragma: no cover
    from functools32 import lru_cache as _lru_cache  # py<3.2


T = typing.TypeVar('T')


class typed_lru_cache(object):

    __slots__ = ('maxsize', 'uncached_function', 'cached_function')

    def __init__(self, maxsize=None):
        # type: (typing.Optional[int]) -> None
        assert isinstance(maxsize, (int, type(None)))
        self.maxsize = maxsize  # type: typing.Optional[int]
        self.uncached_function = None  # type: typing.Optional[typing.Any]
        self.cached_function = None  # type: typing.Optional[typing.Any]

    def __call__(self, fn):
        # type: (T) -> T
        # assert (self.cached_function is not None) == (self.uncached_function is not None)

        if self.cached_function is None:
            self.uncached_function = fn  # type: T
            self.cached_function = _lru_cache(maxsize=self.maxsize)(fn)
        else:
            assert self.uncached_function == fn  # pragma: no cover  # defensive approach, this should not happen

        return typing.cast(T, self.cached_function)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing
from itertools import chain

from bravado_core.spec import Spec

from swagger_spec_compatibility.util import EntityMapping
from swagger_spec_compatibility.walkers import NoValue
from swagger_spec_compatibility.walkers import PathType
from swagger_spec_compatibility.walkers import SchemaWalker


def _different_enum_values_mapping(
    left_spec,  # type: Spec
    right_spec,  # type: Spec
    left_schema,  # type: typing.Optional[typing.Mapping[typing.Text, typing.Any]]
    right_schema,  # type: typing.Optional[typing.Mapping[typing.Text, typing.Any]]
):
    # type: (...) -> typing.Optional[EntityMapping[typing.Set[typing.Text]]]
    left_enum_values = (
        set(left_schema['enum'])
        if (left_schema and left_schema.get('type') == 'string' and 'enum' in left_schema)
        else set()
    )
    right_enum_values = (
        set(right_schema['enum'])
        if (right_schema and right_schema.get('type') == 'string' and 'enum' in right_schema)
        else set()
    )

    enum_values_appear_once = left_enum_values.symmetric_difference(right_enum_values)
    if not enum_values_appear_once:
        # The condition is true if left_required is empty and right_required is not empty or vice-versa
        return None
    else:
        return EntityMapping(
            old=enum_values_appear_once.intersection(left_enum_values),
            new=enum_values_appear_once.intersection(right_enum_values),
        )


class EnumValuesDiff(typing.NamedTuple(
    'EnumValuesDiff', (
        ('path', PathType),
        ('mapping', EntityMapping[typing.Any]),
    ),
)):
    def fix_parameter_path(self, path, original_path):
        # type: (PathType, PathType) -> 'EnumValuesDiff'
        return EnumValuesDiff(
            path=tuple(chain(original_path, self.path[len(original_path):])),
            mapping=self.mapping,
        )


class EnumValuesDifferWalker(SchemaWalker[EnumValuesDiff]):
    left_spec = None  # type: Spec
    right_spec = None  # type: Spec

    def dict_check(
        self,
        path,  # type: PathType
        left_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
        right_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
    ):
        # type: (...) -> typing.Iterable[EnumValuesDiff]
        different_enum_values_mapping = _different_enum_values_mapping(
            left_spec=self.left_spec,
            right_spec=self.right_spec,
            left_schema=None if isinstance(left_dict, NoValue) else left_dict,
            right_schema=None if isinstance(right_dict, NoValue) else right_dict,
        )
        if different_enum_values_mapping is not None:
            return (
                EnumValuesDiff(
                    path=path,
                    mapping=different_enum_values_mapping,
                ),
            )
        else:
            return ()

    def list_check(
        self,
        path,  # type: PathType
        left_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
        right_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
    ):
        # type: (...) -> typing.Iterable[EnumValuesDiff]
        return ()

    def value_check(
        self,
        path,  # type: PathType
        left_value,  # type: typing.Any
        right_value,  # type: typing.Any
    ):
        # type: (...) -> typing.Iterable[EnumValuesDiff]
        return ()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import argparse
from textwrap import dedent

from six import iteritems

from swagger_spec_compatibility.cli import explain
from swagger_spec_compatibility.cli import info
from swagger_spec_compatibility.cli import run
from swagger_spec_compatibility.util import wrap


_SUB_COMMAND_ASSOCIATED_FUNCTION_MAPPING = {
    explain.add_sub_parser: explain.execute,
    info.add_sub_parser: info.execute,
    run.add_sub_parser: run.execute,
}


def parser():
    # type: () -> argparse.ArgumentParser

    argument_parser = argparse.ArgumentParser(
        description=wrap(dedent("""
            Tool for the identification of backward incompatible changes between two swagger specs.

            The tool provides the following level of results:
            - WARNING: the Swagger specs are technically compatible but the are likely to break known Swagger implementations
            - ERROR: new Swagger spec does introduce a breaking change respect the old implementation
        """)),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    subparsers = argument_parser.add_subparsers(help='help for sub-command', dest='command')
    subparsers.required = True
    for add_sub_command_parser, associated_function in iteritems(_SUB_COMMAND_ASSOCIATED_FUNCTION_MAPPING):
        add_sub_command_parser(subparsers).set_defaults(func=associated_function)

    return argument_parser



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing
from textwrap import TextWrapper

from swagger_spec_compatibility.walkers import PathType

T = typing.TypeVar('T')


def wrap(text, width=120, indent=''):
    # type: (typing.Text, int, typing.Text) -> typing.Text
    wrapper = TextWrapper(
        expand_tabs=False,
        replace_whitespace=False,
        break_long_words=False,
        width=width,
        initial_indent=str(indent),
        subsequent_indent=str(indent),
    )
    return '\n'.join('\n'.join(wrapper.wrap(line)) for line in text.splitlines())


class EntityMapping(typing.Generic[T]):
    __slots__ = ('_old', '_new')

    def __init__(self, old, new):
        # type: (T, T) -> None
        self._old = old
        self._new = new

    @property  # Property needed as temporary workaround to generic named tuples (https://github.com/python/mypy/issues/685)
    def old(self):
        # type: () -> T
        return self._old

    @property  # Property needed as temporary workaround to generic named tuples (https://github.com/python/mypy/issues/685)
    def new(self):
        # type: () -> T
        return self._new

    def __iter__(self):
        # type: () -> typing.Generator[T, None, None]
        yield self.old
        yield self.new

    def __hash__(self):
        # type: () -> int
        return hash((hash(self._old), hash(self._new)))

    def __eq__(self, other):
        # type: (typing.Any) -> bool
        return isinstance(other, self.__class__) and self._old == other._old and self._new == other._new

    def __repr__(self):
        # type: () -> str
        return str('{}(old={}, new={})'.format(
            self.__class__.__name__, self.old,
            self.new,
        ))  # pragma: no cover  # This statement is present only to have a nicer REPL experience # noqa


def is_path_in_top_level_paths(top_level_paths, path):
    # type: (typing.Iterable[PathType], PathType) -> bool
    return any(
        path[:len(acceptable_path)] == acceptable_path
        for acceptable_path in top_level_paths
    )



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing

from bravado_core.spec import Spec

from swagger_spec_compatibility.rules.common import BaseRule
from swagger_spec_compatibility.rules.common import Level
from swagger_spec_compatibility.rules.common import RuleType
from swagger_spec_compatibility.rules.common import ValidationMessage
from swagger_spec_compatibility.util import is_path_in_top_level_paths
from swagger_spec_compatibility.walkers import format_path
from swagger_spec_compatibility.walkers.enum_values import EnumValuesDifferWalker
from swagger_spec_compatibility.walkers.request_parameters import RequestParametersWalker


class RemovedEnumValueFromRequest(BaseRule):
    description = 'Removing an enum value from a request parameter is backward incompatible as a previously valid ' \
                  'request will not be valid. This happens because a request containing the removed enum value, ' \
                  'valid according to the "old" Swagger spec, is not valid according to the new specs.'
    error_code = 'REQ-E002'
    error_level = Level.ERROR
    rule_type = RuleType.REQUEST_CONTRACT
    short_name = 'Removed Enum value from Request contract'

    @classmethod
    def validate(cls, left_spec, right_spec):
        # type: (Spec, Spec) -> typing.Iterable[ValidationMessage]
        request_parameters_paths = RequestParametersWalker(left_spec, right_spec).walk()

        # FIXME: the used walker is not able to merge together parameters defined in different locations
        for enum_values_diff in EnumValuesDifferWalker(left_spec, right_spec).walk():
            if not enum_values_diff.mapping.old:
                continue
            if not is_path_in_top_level_paths(request_parameters_paths, enum_values_diff.path):
                continue
            yield cls.validation_message(format_path(enum_values_diff.path))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import glob
import os
import re
from typing import Any
from typing import Generator
from typing import NamedTuple
from typing import Text

from six import itervalues

from swagger_spec_compatibility.cli.common import uri
from swagger_spec_compatibility.rules import compatibility_status
from swagger_spec_compatibility.spec_utils import load_spec_from_uri


Specification = NamedTuple(
    'Specification', [
        ('test_id', Text),
        ('old_spec_uri', Text),
        ('new_spec_uri', Text),
        ('number_of_reports', int),
    ],
)


def _test_specification_id(test_specification):  # pragma: no cover # this is used internally by pytest and it won't appear on coverage
    # type: (Specification) -> Text
    return test_specification.test_id


def get_test_specifications():  # pragma: no cover # this is used internally by pytest and it won't appear on coverage
    # type: () -> Generator[Specification, None, None]
    test_specs_dir = os.path.join(os.path.dirname(__file__), 'test-specs')
    for case_dir in glob.iglob(os.path.join(test_specs_dir, 'case-*-*-reports*')):
        match = re.match(r'^.*/case-\d+-(?P<number_of_reports>\d+)-reports.*$', case_dir)
        if not match:
            raise RuntimeError('{} does not follow the expected pattern (case-\\d+-\\d+-reports(-description)?'.format(case_dir))

        yield Specification(
            test_id=os.path.basename(case_dir),
            old_spec_uri=uri(str(os.path.join(case_dir, 'old.yaml'))),
            new_spec_uri=uri(str(os.path.join(case_dir, 'new.yaml'))),
            number_of_reports=int(match.group(str('number_of_reports'))),
        )


def pytest_generate_tests(metafunc):  # pragma: no cover # this is used internally by pytest and it won't appear on coverage
    # type: (Any) -> None
    if metafunc.definition.name == 'test_spec_from_test_specs_directory' and ['test_specification'] == metafunc.fixturenames:
        test_specifications = list(get_test_specifications())
        if test_specifications:
            metafunc.parametrize('test_specification', test_specifications, ids=_test_specification_id)
        else:
            metafunc.parametrize('test_specification', [])


def test_spec_from_test_specs_directory(test_specification):
    # type: (Specification) -> None
    result = compatibility_status(
        old_spec=load_spec_from_uri(test_specification.old_spec_uri),
        new_spec=load_spec_from_uri(test_specification.new_spec_uri),
    )
    reports = [
        report
        for reports in itervalues(result)
        for report in reports
    ]
    error_message = '{} reports were expected but {} reports received\n{}'.format(
        test_specification.number_of_reports,
        len(reports),
        '\n'.join(report.string_representation() for report in reports),
    )
    assert len(reports) == test_specification.number_of_reports, error_message



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing

from swagger_spec_compatibility.cli import parser
from swagger_spec_compatibility.cli.common import post_process_rules_cli_arguments
from swagger_spec_compatibility.cli.common import pre_process_cli_to_discover_rules


def main(argv=None):
    # type: (typing.Optional[typing.Sequence[typing.Text]]) -> int
    pre_process_cli_to_discover_rules(argv)
    args = post_process_rules_cli_arguments(parser().parse_args(argv))
    exit_code = args.func(args)  # type: int
    return exit_code


if __name__ == '__main__':
    exit(main())



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing

from bravado_core.spec import Spec

from swagger_spec_compatibility.rules.common import BaseRule
from swagger_spec_compatibility.rules.common import Level
from swagger_spec_compatibility.rules.common import RuleType
from swagger_spec_compatibility.rules.common import ValidationMessage
from swagger_spec_compatibility.util import is_path_in_top_level_paths
from swagger_spec_compatibility.walkers import format_path
from swagger_spec_compatibility.walkers.changed_types import ChangedTypesDifferWalker
from swagger_spec_compatibility.walkers.request_parameters import RequestParametersWalker
from swagger_spec_compatibility.walkers.response_paths import ResponsePathsWalker


class ChangedType(BaseRule):
    description = 'Changing the type of a field is not backward compatible as a client using "old" Swagger specs will ' \
                  'send the field with a different type leading the service to fail to validate the request. ' \
                  'On the other end, if the object containing the updated field is used in the response, ' \
                  'it will lead to unexpected client errors when parsing the response and/or using ' \
                  'the updated property.'
    error_code = 'MIS-E002'
    error_level = Level.ERROR
    rule_type = RuleType.MISCELLANEOUS
    short_name = 'Changed type'

    @classmethod
    def validate(cls, left_spec, right_spec):
        # type: (Spec, Spec) -> typing.Iterable[ValidationMessage]

        request_parameter_paths = RequestParametersWalker(left_spec, right_spec).walk()
        response_paths = ResponsePathsWalker(left_spec, right_spec).walk()

        for changed_types_diff in ChangedTypesDifferWalker(left_spec, right_spec).walk():
            if (
                not is_path_in_top_level_paths(request_parameter_paths, changed_types_diff.path) and
                not is_path_in_top_level_paths(response_paths, changed_types_diff.path)
            ):
                continue
            yield cls.validation_message(format_path(changed_types_diff.path))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing
from itertools import chain

from bravado_core.spec import Spec

from swagger_spec_compatibility.util import EntityMapping
from swagger_spec_compatibility.walkers import NoValue
from swagger_spec_compatibility.walkers import PathType
from swagger_spec_compatibility.walkers import SchemaWalker


def _different_types_mapping(
    left_spec,  # type: Spec
    right_spec,  # type: Spec
    left_schema,  # type: typing.Optional[typing.Mapping[typing.Text, typing.Any]]
    right_schema,  # type: typing.Optional[typing.Mapping[typing.Text, typing.Any]]
):
    # type: (...) -> typing.Optional[EntityMapping[typing.Optional[typing.Text]]]
    left_type = left_schema.get('type') if left_schema is not None else None
    right_type = right_schema.get('type') if right_schema is not None else None

    if not left_type and left_spec.config['default_type_to_object']:  # Normalize type according to spec configuration
        left_type = 'object'
    if not right_type and right_spec.config['default_type_to_object']:  # Normalize type according to spec configuration
        right_type = 'object'

    if left_type == right_type:
        return None
    else:
        return EntityMapping(
            old=left_type,
            new=right_type,
        )


class ChangedTypesDiff(typing.NamedTuple(
    'ChangedTypesDiff', (
        ('path', PathType),
        ('mapping', EntityMapping[typing.Optional[typing.Text]]),
    ),
)):
    def fix_parameter_path(self, path, original_path):
        # type: (PathType, PathType) -> 'ChangedTypesDiff'
        return ChangedTypesDiff(
            path=tuple(chain(original_path, self.path[len(original_path):])),
            mapping=self.mapping,
        )


class ChangedTypesDifferWalker(SchemaWalker[ChangedTypesDiff]):
    left_spec = None  # type: Spec
    right_spec = None  # type: Spec
    diffs = None  # type: typing.List[ChangedTypesDiff]

    def __init__(
        self,
        left_spec,  # type: Spec
        right_spec,  # type: Spec
    ):
        # type: (...) -> None
        super(ChangedTypesDifferWalker, self).__init__(
            left_spec=left_spec,
            right_spec=right_spec,
        )

    def dict_check(
        self,
        path,  # type: PathType
        left_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
        right_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
    ):
        # type: (...) -> typing.Iterable[ChangedTypesDiff]
        different_types_mapping = _different_types_mapping(
            left_spec=self.left_spec,
            right_spec=self.right_spec,
            left_schema=None if isinstance(left_dict, NoValue) else left_dict,
            right_schema=None if isinstance(right_dict, NoValue) else right_dict,
        )
        if different_types_mapping:
            return (
                ChangedTypesDiff(
                    path=path,
                    mapping=different_types_mapping,
                ),
            )
        else:
            return ()

    def list_check(
        self,
        path,  # type: PathType
        left_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
        right_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
    ):
        # type: (...) -> typing.Iterable[ChangedTypesDiff]
        return ()

    def value_check(
        self,
        path,  # type: PathType
        left_value,  # type: typing.Any
        right_value,  # type: typing.Any
    ):
        # type: (...) -> typing.Iterable[ChangedTypesDiff]
        return ()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import json
import sys
import typing

from six import iteritems

from swagger_spec_compatibility.cli.common import add_rules_arguments
from swagger_spec_compatibility.cli.common import CLIProtocol
from swagger_spec_compatibility.cli.common import rules
from swagger_spec_compatibility.cli.common import uri
from swagger_spec_compatibility.rules import compatibility_status
from swagger_spec_compatibility.rules import ValidationMessage
from swagger_spec_compatibility.rules.common import Level
from swagger_spec_compatibility.rules.common import RuleProtocol
from swagger_spec_compatibility.spec_utils import load_spec_from_uri


class _Namespace(CLIProtocol):
    blacklist_rules = None  # type: typing.Iterable[typing.Text]
    rules = None  # type: typing.Iterable[typing.Text]
    strict = None  # type: bool
    json_output = None  # type: bool
    old_spec = None  # type: typing.Text
    new_spec = None  # type: typing.Text


def _extract_rules_with_given_message_level(
    rules_to_messages_mapping,  # type: typing.Mapping[typing.Type[RuleProtocol], typing.Iterable[ValidationMessage]]
    level,  # type: Level
):
    # type: (...) -> typing.List[ValidationMessage]
    return [
        message
        for rule, messages in iteritems(rules_to_messages_mapping)
        for message in messages
        if message.level is level
    ]


def _print_raw_messages(messages_by_level):
    # type: (typing.Mapping[Level, typing.Iterable[ValidationMessage]]) -> None
    for level, messages in iteritems(messages_by_level):
        messages = list(messages)
        if messages:
            print(
                '{} rules:\n\t{}'.format(
                    level.name,
                    '\n\t'.join(
                        message.string_representation()
                        for message in messages
                    ),
                ),
            )


def _print_json_messages(messages_by_level):
    # type: (typing.Mapping[Level, typing.Iterable[ValidationMessage]]) -> None
    json_output = {}
    for level, messages in iteritems(messages_by_level):
        level_output = [
            message.json_representation()
            for message in messages
        ]
        if level_output:
            json_output[level.name] = level_output
    json.dump(json_output, sys.stdout)


def execute(cli_args):
    # type: (_Namespace) -> int
    rules_to_messages_mapping = compatibility_status(
        old_spec=load_spec_from_uri(cli_args.old_spec),
        new_spec=load_spec_from_uri(cli_args.new_spec),
        rules=rules(cli_args),
    )

    messages_by_level = {
        level: _extract_rules_with_given_message_level(rules_to_messages_mapping, level)
        for level in Level
    }

    if cli_args.json_output:
        _print_json_messages(messages_by_level)
    else:
        _print_raw_messages(messages_by_level)

    if cli_args.strict:
        return 1 if any(messages_by_level.values()) else 0
    else:
        return 1 if any(messages_by_level[Level.ERROR]) else 0


def add_sub_parser(subparsers):
    # type: (argparse._SubParsersAction) -> argparse.ArgumentParser
    run_detection_parser = subparsers.add_parser('run', help='run backward compatibility detection')

    add_rules_arguments(run_detection_parser)

    run_detection_parser.add_argument(
        '--strict',
        action='store_true',
        help='Convert warnings to errors',
    )
    run_detection_parser.add_argument(
        '--json-output',
        action='store_true',
        help='Return machine readable json output',
    )
    run_detection_parser.add_argument(
        'old_spec',
        type=uri,
        help='Path/URI of the "old" version of the Swagger spec',
    )
    run_detection_parser.add_argument(
        'new_spec',
        type=uri,
        help='Path/URI of the "new" version of the Swagger spec',
    )
    return run_detection_parser



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing
from itertools import chain

from bravado_core.spec import Spec

from swagger_spec_compatibility.spec_utils import get_required_properties
from swagger_spec_compatibility.util import EntityMapping
from swagger_spec_compatibility.walkers import NoValue
from swagger_spec_compatibility.walkers import PathType
from swagger_spec_compatibility.walkers import SchemaWalker


def _different_properties_mapping(
    left_spec,  # type: Spec
    right_spec,  # type: Spec
    left_schema,  # type: typing.Optional[typing.Mapping[typing.Text, typing.Any]]
    right_schema,  # type: typing.Optional[typing.Mapping[typing.Text, typing.Any]]
):
    # type: (...) -> typing.Optional[EntityMapping[typing.Set[typing.Text]]]
    left_required = get_required_properties(swagger_spec=left_spec, schema=left_schema) or set()
    right_required = get_required_properties(swagger_spec=right_spec, schema=right_schema) or set()

    properties_appear_once = left_required.symmetric_difference(right_required)
    if not properties_appear_once:
        # The condition is true if left_required is empty and right_required is not empty or vice-versa
        return None
    else:
        return EntityMapping(
            old=properties_appear_once.intersection(left_required),
            new=properties_appear_once.intersection(right_required),
        )


class RequiredPropertiesDiff(typing.NamedTuple(
    'RequiredPropertiesDiff', (
        ('path', PathType),
        ('mapping', EntityMapping[typing.Set[typing.Text]]),
    ),
)):
    def fix_parameter_path(self, path, original_path):
        # type: (PathType, PathType) -> 'RequiredPropertiesDiff'
        return RequiredPropertiesDiff(
            path=tuple(chain(original_path, self.path[len(original_path):])),
            mapping=self.mapping,
        )


class RequiredPropertiesDifferWalker(SchemaWalker[RequiredPropertiesDiff]):
    left_spec = None  # type: Spec
    right_spec = None  # type: Spec
    diffs = None  # type: typing.List[RequiredPropertiesDiff]

    def __init__(
        self,
        left_spec,  # type: Spec
        right_spec,  # type: Spec
    ):
        # type: (...) -> None
        super(RequiredPropertiesDifferWalker, self).__init__(
            left_spec=left_spec,
            right_spec=right_spec,
        )

    def dict_check(
        self,
        path,  # type: PathType
        left_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
        right_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
    ):
        # type: (...) -> typing.Iterable[RequiredPropertiesDiff]
        different_properties_mapping = _different_properties_mapping(
            left_spec=self.left_spec,
            right_spec=self.right_spec,
            left_schema=None if isinstance(left_dict, NoValue) else left_dict,
            right_schema=None if isinstance(right_dict, NoValue) else right_dict,
        )
        if different_properties_mapping:
            return (
                RequiredPropertiesDiff(
                    path=path,
                    mapping=different_properties_mapping,
                ),
            )
        else:
            return ()

    def list_check(
        self,
        path,  # type: PathType
        left_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
        right_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
    ):
        # type: (...) -> typing.Iterable[RequiredPropertiesDiff]
        return ()

    def value_check(
        self,
        path,  # type: PathType
        left_value,  # type: typing.Any
        right_value,  # type: typing.Any
    ):
        # type: (...) -> typing.Iterable[RequiredPropertiesDiff]
        return ()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing
from abc import ABCMeta
from abc import abstractmethod
from enum import IntEnum

import typing_extensions
from bravado_core.spec import Spec
from six import iterkeys
from six import itervalues
from six import with_metaclass
from termcolor import colored

from swagger_spec_compatibility.util import wrap


def _read_the_docs_link(rule):
    # type: (typing.Type['RuleProtocol']) -> typing.Text
    return 'https://swagger-spec-compatibility.readthedocs.io/en/latest/rules/{code}.html'.format(
        code=rule.error_code,
    )


def get_rule_documentation_link(rule):
    # type: (typing.Type['RuleProtocol']) -> typing.Optional[typing.Text]
    """
    Helper method that allows to extract documentation link related to a given rule.

    If the rule is implemented within swagger-spec-compatibility library then the documentation
    link will fall back to the "default" read-the-docs link
    """
    if rule.documentation_link:
        return rule.documentation_link
    elif rule.__module__.startswith('swagger_spec_compatibility.rules'):
        return _read_the_docs_link(rule)
    else:
        return None


class RuleRegistry(ABCMeta):
    _REGISTRY = {}  # type: typing.MutableMapping[typing.Text, typing.Type['BaseRule']]

    @classmethod
    def _validate_class_attributes(mcs, cls):
        # type: (typing.Type['BaseRule']) -> None
        assert getattr(cls, 'error_code', None) is not None, 'error_code is a required class attribute for {}'.format(cls)
        assert getattr(cls, 'short_name', None) is not None, 'short_name is a required class attribute for {}'.format(cls)
        assert getattr(cls, 'description', None) is not None, 'description is a required class attribute for {}'.format(cls)
        assert getattr(cls, 'error_level', None) is not None, 'error_level is a required class attribute for {}'.format(cls)
        assert getattr(cls, 'rule_type', None) is not None, 'rule_type is a required class attribute for {}'.format(cls)

    @classmethod
    def _prevent_rule_duplication(mcs, cls):
        # type: (typing.Type['BaseRule']) -> None
        assert cls.error_code not in RuleRegistry._REGISTRY, \
            'Rule {} is already defined. Already existing: {} Trying to add: {}'.format(
                cls.error_code,
                RuleRegistry._REGISTRY[cls.error_code],
                cls,
            )

    @classmethod
    def _prevent_metaclass_usage_from_not_BaseRule_extensions(mcs, cls):
        # type: (typing.Type) -> None
        cls_fully_qualified_name = '{}.{}'.format(cls.__module__, cls.__name__)
        base_rule_fully_qualified_name = '{}.BaseRule'.format(mcs.__module__)
        assert (  # pragma: no branch
            cls_fully_qualified_name == base_rule_fully_qualified_name
            or any(base_rule_fully_qualified_name == '{}.{}'.format(c.__module__, c.__name__) for c in cls.__bases__)
        ), '{} metaclass should be used only by {}.BaseRule'.format(mcs, mcs.__module__)

    def __new__(mcs, name, bases, namespace):
        # type: (typing.Type['RuleRegistry'], str, typing.Tuple[type, ...], typing.Dict[str, typing.Any]) -> type
        new_cls = ABCMeta.__new__(mcs, name, bases, namespace)  # type: typing.Type['BaseRule']
        mcs._prevent_metaclass_usage_from_not_BaseRule_extensions(new_cls)

        if name != 'BaseRule':
            mcs._validate_class_attributes(new_cls)
            mcs._prevent_rule_duplication(new_cls)
            RuleRegistry._REGISTRY[new_cls.error_code] = new_cls
        return new_cls

    @staticmethod
    def rule_names():
        # type: () -> typing.Iterable[typing.Text]
        return sorted(iterkeys(RuleRegistry._REGISTRY))

    @staticmethod
    def rules():
        # type: () -> typing.Iterable[typing.Type['BaseRule']]
        return sorted(itervalues(RuleRegistry._REGISTRY), key=lambda rule: rule.error_code)

    @staticmethod
    def has_rule(rule_name):
        # type: (typing.Text) -> bool
        return rule_name in RuleRegistry._REGISTRY

    @staticmethod
    def rule(rule_name):
        # type: (typing.Text) -> typing.Type['BaseRule']
        return RuleRegistry._REGISTRY[rule_name]


class RuleType(IntEnum):
    REQUEST_CONTRACT = 0
    RESPONSE_CONTRACT = 1
    MISCELLANEOUS = 2


class Level(IntEnum):
    INFO = 0
    WARNING = 1
    ERROR = 2


class RuleProtocol(typing_extensions.Protocol):
    # Unique identifier of the rule
    error_code = None  # type: typing_extensions.ClassVar[typing.Text]
    # Short name of the rule. This will be visible on CLI in case the rule is triggered
    short_name = None  # type: typing_extensions.ClassVar[typing.Text]
    # Short description of the rationale of the rule. This will be visible on CLI only.
    description = None  # type: typing_extensions.ClassVar[typing.Text]
    # Error level associated to the rule
    error_level = None  # type: typing_extensions.ClassVar[Level]
    # Type of the rule associated
    rule_type = None  # type: typing_extensions.ClassVar[RuleType]
    # Documentation link
    documentation_link = None  # type: typing_extensions.ClassVar[typing.Optional[typing.Text]]

    @classmethod
    def validate(cls, left_spec, right_spec):
        # type: (Spec, Spec) -> typing.Iterable['ValidationMessage']
        pass


class ValidationMessage(typing.NamedTuple(
    '_ValidationMessage', (
        ('level', Level),
        ('rule', typing.Type[RuleProtocol]),
        ('reference', typing.Text),
    ),
)):
    def string_representation(self):
        # type: () -> typing.Text
        documentation_link = get_rule_documentation_link(self.rule)
        return '[{error_code}] {short_name}: {reference}{more_info}'.format(
            error_code=self.rule.error_code,
            reference=self.reference,
            short_name=self.rule.short_name,
            more_info=' (documentation: {})'.format(documentation_link) if documentation_link else '',
        )

    def json_representation(self):
        # type: () -> typing.Mapping[typing.Text, typing.Any]
        return {
            'error_code': self.rule.error_code,
            'reference': self.reference,
            'short_name': self.rule.short_name,
            'documentation': get_rule_documentation_link(self.rule),
        }


class BaseRule(with_metaclass(RuleRegistry)):
    # Unique identifier of the rule
    error_code = None  # type: typing_extensions.ClassVar[typing.Text]
    # Short name of the rule. This will be visible on CLI in case the rule is triggered
    short_name = None  # type: typing_extensions.ClassVar[typing.Text]
    # Short description of the rationale of the rule. This will be visible on CLI only.
    description = None  # type: typing_extensions.ClassVar[typing.Text]
    # Error level associated to the rule
    error_level = None  # type: typing_extensions.ClassVar[Level]
    # Type of the rule associated
    rule_type = None  # type: typing_extensions.ClassVar[RuleType]
    # Documentation link
    documentation_link = None  # type: typing_extensions.ClassVar[typing.Optional[typing.Text]]

    def __init__(self):
        # type: () -> None
        raise RuntimeError('This class should not be initialized. The assumed usage is via class methods.')

    @classmethod
    @abstractmethod
    def validate(cls, left_spec, right_spec):
        # type: (Spec, Spec) -> typing.Iterable[ValidationMessage]
        pass

    @classmethod
    def explain(cls):
        # type: () -> typing.Text
        documentation_link = get_rule_documentation_link(cls)
        return '[{error_code}] {short_name}:\n{rule_description}{more_info}'.format(
            error_code=colored(cls.error_code, attrs=['bold']),
            short_name=colored(cls.short_name, color='cyan', attrs=['bold']),
            rule_description=wrap(cls.description, indent='\t'),
            more_info='\n\nMore info on {}'.format(documentation_link) if documentation_link else '',
        )

    @classmethod
    def validation_message(cls, reference):
        # type: (typing.Text) -> ValidationMessage
        return ValidationMessage(
            level=cls.error_level,
            rule=cls,
            reference=reference,
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing
from enum import Enum
from itertools import chain

from bravado.client import SwaggerClient
from bravado_core.operation import Operation
from bravado_core.spec import Spec
from bravado_core.util import determine_object_type
from bravado_core.util import ObjectType
from six import iterkeys
from swagger_spec_validator.validator20 import get_collapsed_properties_type_mappings

from swagger_spec_compatibility.cache import typed_lru_cache
from swagger_spec_compatibility.util import EntityMapping


class HTTPVerb(Enum):
    DELETE = 'delete'
    GET = 'get'
    HEAD = 'head'
    OPTIONS = 'options'
    PARAMETERS = 'parameters'
    PATCH = 'patch'
    POST = 'post'
    PUT = 'put'

    @staticmethod
    def from_swagger_operation(operation):
        # type: (Operation) -> 'HTTPVerb'
        return HTTPVerb(operation.http_method)


class Endpoint(typing.NamedTuple(
    '_Endpoint', (
        ('http_verb', HTTPVerb),
        ('path', typing.Text),
        ('operation', Operation),
    ),
)):

    @staticmethod
    def from_swagger_operation(operation):
        # type: (Operation) -> 'Endpoint'
        return Endpoint(
            http_verb=HTTPVerb.from_swagger_operation(operation),
            path=operation.path_name,
            operation=operation,
        )

    def __hash__(self):
        # type: () -> int
        return hash((hash(self.http_verb), hash(self.path)))

    def __eq__(self, other):
        # type: (typing.Any) -> bool
        return isinstance(other, self.__class__) and self.http_verb == other.http_verb and self.path == other.path


@typed_lru_cache(maxsize=2)
def load_spec_from_uri(uri):
    # type: (typing.Text) -> Spec
    return SwaggerClient.from_url(uri, config={'internally_dereference_refs': True}).swagger_spec


def load_spec_from_spec_dict(spec_dict):
    # type: (typing.Mapping[typing.Text, typing.Any]) -> Spec
    return SwaggerClient.from_spec(spec_dict, origin_url='', config={'internally_dereference_refs': True}).swagger_spec


@typed_lru_cache(maxsize=2)
def get_operations(spec):
    # type: (Spec) -> typing.List[Operation]
    return [
        operation
        for resource in spec.resources.values()
        for operation in resource.operations.values()
    ]


@typed_lru_cache(maxsize=2)
def get_endpoints(spec):
    # type: (Spec) -> typing.Set[Endpoint]
    return {
        Endpoint.from_swagger_operation(operation)
        for operation in get_operations(spec)
    }


def get_operation_mappings(old_spec, new_spec):
    # type: (Spec, Spec) -> typing.Set[EntityMapping[Operation]]
    old_endpoints = get_endpoints(old_spec)
    old_endpoints_map = {  # Small hack to make endpoint search easy
        endpoint: endpoint
        for endpoint in old_endpoints
    }
    new_endpoints = get_endpoints(new_spec)
    new_endpoints_map = {  # Small hack to make endpoint search easy
        endpoint: endpoint
        for endpoint in new_endpoints
    }

    return {
        EntityMapping(old_endpoints_map[endpoint].operation, new_endpoints_map[endpoint].operation)
        for endpoint in old_endpoints.intersection(new_endpoints)
    }


def get_required_properties(swagger_spec, schema):
    # type: (Spec, typing.Optional[typing.Mapping[typing.Text, typing.Any]]) -> typing.Optional[typing.Set[typing.Text]]
    if schema is None or determine_object_type(schema) != ObjectType.SCHEMA:
        return None
    required, _ = get_collapsed_properties_type_mappings(definition=schema, deref=swagger_spec.deref)
    return set(iterkeys(required))


def get_properties(swagger_spec, schema):
    # type: (Spec, typing.Optional[typing.Mapping[typing.Text, typing.Any]]) -> typing.Optional[typing.Set[typing.Text]]
    if schema is None or determine_object_type(schema) != ObjectType.SCHEMA:
        return None
    required, not_required = get_collapsed_properties_type_mappings(definition=schema, deref=swagger_spec.deref)
    return set(chain(iterkeys(required), iterkeys(not_required)))


StatusCodeSchema = typing.NamedTuple(
    'StatusCodeSchema', (
        ('status_code', typing.Text),
        ('mapping', EntityMapping[typing.Optional[typing.Mapping[typing.Text, typing.Any]]]),
    ),
)


def iterate_on_responses_status_codes(
    old_operation,  # type: typing.Mapping[typing.Text, typing.Any]
    new_operation,  # type: typing.Mapping[typing.Text, typing.Any]
):
    # type: (...) -> typing.Generator[StatusCodeSchema, None, None]
    old_status_code_schema_mapping = old_operation.get('responses') or {}
    new_status_code_schema_mapping = new_operation.get('responses') or {}

    common_response_codes = set(iterkeys(old_status_code_schema_mapping)).intersection(
        set(iterkeys(new_status_code_schema_mapping)),
    )
    # Compare schemas for the same status code only (TODO: what to do for old=default and new=404?)
    for status_code in common_response_codes:
        yield StatusCodeSchema(
            status_code=status_code,
            mapping=EntityMapping(
                old=old_status_code_schema_mapping[status_code].get('schema'),
                new=new_status_code_schema_mapping[status_code].get('schema'),
            ),
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import venusian

import swagger_spec_compatibility.rules

# Scan the whole library in order to import all the available rules
venusian.Scanner().scan(swagger_spec_compatibility.rules)



# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup

with open('README.rst') as readme:
    long_description = readme.read()

setup(
    name='swagger-spec-compatibility',
    version='1.2.1',
    author='Yelp, Inc.',
    author_email='opensource+swagger-spec-compatibility@yelp.com',
    description='Python library to check Swagger Spec backward compatibility',
    long_description=long_description,
    url='https://github.com/Yelp/swagger-spec-compatibility',
    install_requires=[
        'bravado',
        'bravado-core != 5.16.0',
        'swagger-spec-validator',
        'six',
        'termcolor',
        'typing_extensions',
        'venusian',
    ],
    extras_require={
        ':python_version<"3.5"': ['typing'],
        ':python_version<"3.2"': ['functools32'],
    },
    license='Copyright Yelp, Inc. 2018',
    packages=find_packages(exclude=('tests*', 'testing*')),
    package_data={
        'swagger_spec_compatibility': ['py.typed'],
    },
    entry_points={
        'console_scripts': [
            'swagger_spec_compatibility = swagger_spec_compatibility.__main__:main',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing
from enum import Enum
from itertools import chain

from bravado_core.spec import Spec

from swagger_spec_compatibility.spec_utils import get_properties
from swagger_spec_compatibility.util import EntityMapping
from swagger_spec_compatibility.walkers import NoValue
from swagger_spec_compatibility.walkers import PathType
from swagger_spec_compatibility.walkers import SchemaWalker


class DiffType(Enum):
    PROPERTIES = 0
    VALUE = 1


class AdditionalPropertiesDiff(typing.NamedTuple(
    'AdditionalPropertiesDiff', (
        ('path', PathType),
        ('diff_type', DiffType),
        ('additionalProperties', typing.Optional[EntityMapping[typing.Union[bool, typing.Mapping[typing.Text, typing.Any]]]]),
        ('properties', typing.Optional[EntityMapping[typing.Set[typing.Text]]]),
    ),
)):
    def fix_parameter_path(self, path, original_path):
        # type: (PathType, PathType) -> 'AdditionalPropertiesDiff'
        return AdditionalPropertiesDiff(
            path=tuple(chain(original_path, self.path[len(original_path):])),
            diff_type=self.diff_type,
            additionalProperties=self.additionalProperties,
            properties=self.properties,
        )


def _evaluate_additional_properties_diffs(
    path,  # type: PathType
    left_spec,  # type: Spec
    right_spec,  # type: Spec
    left_schema,  # type: typing.Optional[typing.Mapping[typing.Text, typing.Any]]
    right_schema,  # type: typing.Optional[typing.Mapping[typing.Text, typing.Any]]
):
    # type: (...) -> typing.List[AdditionalPropertiesDiff]
    result = []  # type: typing.List[AdditionalPropertiesDiff]
    left_additional_properties = None if left_schema is None else left_schema.get('additionalProperties', True)
    right_additional_properties = None if right_schema is None else right_schema.get('additionalProperties', True)
    left_properties = get_properties(left_spec, left_schema) or set()
    right_properties = get_properties(right_spec, right_schema) or set()
    properties_appear_once = left_properties.symmetric_difference(right_properties)

    if left_additional_properties == {}:  # Normalize additional properties
        left_additional_properties = True
    if right_additional_properties == {}:  # Normalize additional properties
        right_additional_properties = True

    if left_additional_properties != right_additional_properties:
        result.append(AdditionalPropertiesDiff(
            path=path,
            diff_type=DiffType.VALUE,
            additionalProperties=EntityMapping(
                # casting here is safe as swagger specs are assumed to be valid and bool or dict are the only possible types
                old=typing.cast(typing.Union[bool, typing.Mapping[typing.Text, typing.Any]], left_additional_properties),
                new=typing.cast(typing.Union[bool, typing.Mapping[typing.Text, typing.Any]], right_additional_properties),
            ),
            properties=None,
        ))

    if (left_additional_properties is False or right_additional_properties is False) and properties_appear_once:
        result.append(AdditionalPropertiesDiff(
            path=path,
            diff_type=DiffType.PROPERTIES,
            additionalProperties=None,
            properties=EntityMapping(
                old=properties_appear_once.intersection(left_properties),
                new=properties_appear_once.intersection(right_properties),
            ),
        ))

    return result


class AdditionalPropertiesDifferWalker(SchemaWalker[AdditionalPropertiesDiff]):
    left_spec = None  # type: Spec
    right_spec = None  # type: Spec
    additionalPropertiesValue = None  # type: typing.Union[bool, typing.Mapping[typing.Text, typing.Any]]
    diffs = None  # type: typing.List[AdditionalPropertiesDiff]

    def dict_check(
        self,
        path,  # type: PathType
        left_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
        right_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
    ):
        # type: (...) -> typing.Iterable[AdditionalPropertiesDiff]

        return _evaluate_additional_properties_diffs(
            path=path,
            left_spec=self.left_spec,
            right_spec=self.right_spec,
            left_schema=None if isinstance(left_dict, NoValue) else left_dict,
            right_schema=None if isinstance(right_dict, NoValue) else right_dict,
        )

    def list_check(
        self,
        path,  # type: PathType
        left_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
        right_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
    ):
        # type: (...) -> typing.Iterable[AdditionalPropertiesDiff]
        return ()

    def value_check(
        self,
        path,  # type: PathType
        left_value,  # type: typing.Any
        right_value,  # type: typing.Any
    ):
        # type: (...) -> typing.Iterable[AdditionalPropertiesDiff]
        return ()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing

from bravado_core.spec import Spec
from bravado_core.util import determine_object_type
from bravado_core.util import ObjectType

from swagger_spec_compatibility.walkers import NoValue
from swagger_spec_compatibility.walkers import PathType
from swagger_spec_compatibility.walkers import SchemaWalker


class ResponsePathsWalker(SchemaWalker[PathType]):
    left_spec = None  # type: Spec
    right_spec = None  # type: Spec
    paths = None  # type: typing.Set[PathType]

    def should_path_be_walked_through(self, path):
        # type: (PathType) -> bool
        if not path:
            return True
        if path[0] != 'paths':
            return False
        if len(path) >= 4:
            # A valid path looks like ('paths', endpoint, http_verb, 'responses')
            return path[3] == 'responses'
        return True

    def __init__(
        self,
        left_spec,  # type: Spec
        right_spec,  # type: Spec
    ):
        # type: (...) -> None
        super(ResponsePathsWalker, self).__init__(left_spec=left_spec, right_spec=right_spec)
        self.paths = set()

    def dict_check(
        self,
        path,  # type: PathType
        left_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
        right_dict,  # type: typing.Union[NoValue, typing.Mapping[typing.Text, typing.Any]]
    ):
        # type: (...) -> typing.Iterable[PathType]
        if determine_object_type(left_dict) == ObjectType.RESPONSE or determine_object_type(right_dict) == ObjectType.RESPONSE:
            return (path, )
        else:
            return ()

    def list_check(
        self,
        path,  # type: PathType
        left_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
        right_list,  # type: typing.Union[NoValue, typing.Sequence[typing.Any]]
    ):
        # type: (...) -> typing.Iterable[PathType]
        return ()

    def value_check(
        self,
        path,  # type: PathType
        left_value,  # type: typing.Any
        right_value,  # type: typing.Any
    ):
        # type: (...) -> typing.Iterable[PathType]
        return ()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import typing
import warnings
from abc import abstractmethod
from collections import defaultdict
from itertools import chain

from bravado_core.spec import Spec
from six import iteritems
from six import iterkeys
from six import text_type
from six.moves import zip_longest


class NoValue(object):
    pass


# HTTP verbs as described by https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#parameterObject
_HTTP_OPERATIONS = {'get', 'put', 'post', 'delete', 'options', 'head', 'patch'}
T = typing.TypeVar('T')
PathType = typing.Tuple[typing.Union[typing.Text, int], ...]
NO_VALUE = NoValue()


def format_path(path):
    # type: (PathType) -> typing.Text
    return '#/{}'.format('/'.join(text_type(path_item) for path_item in path))


class Walker(typing.Generic[T]):
    """
    Generic Walker over two objects.

    The abstract class strips away the details related to dictionary vs list iterations,
    path update etc.
    """

    def __init__(self, left, right, **kwargs):
        # type: (typing.Any, typing.Any, typing.Any) -> None
        self.left = left
        self.right = right
        self._walk_result = NO_VALUE  # type: typing.Union[NoValue, typing.Iterable[T]]
        self._inner_walk_calls = defaultdict(list)  # type: typing.DefaultDict[typing.Tuple[int, ...], typing.List[PathType]]
        for attr_name, attr_value in iteritems(kwargs):
            setattr(self, attr_name, attr_value)

    def should_path_be_walked_through(self, path):
        # type: (PathType) -> bool
        """
        Determine whether to traverse or interrupt traversal of a given path.

        This method allows Walkers to skip traversal of area of the specs that are "not interesting".
        This will allow to write simpler methods and to avoid needless traversing of the Specs.
        """
        return True

    @abstractmethod
    def dict_check(
        self,
        path,  # type: PathType
        left_dict,  # type: typing.Union[typing.Mapping[typing.Text, typing.Any], NoValue]
        right_dict,  # type: typing.Union[typing.Mapping[typing.Text, typing.Any], NoValue]
    ):
        # type: (...) -> typing.Iterable[T]
        """
        Compare the left and right content of path in case both objects are dictionaries.
        """
        raise NotImplementedError()

    @abstractmethod
    def list_check(
        self,
        path,  # type: PathType
        left_list,  # type: typing.Union[typing.Sequence[typing.Any], NoValue]
        right_list,  # type: typing.Union[typing.Sequence[typing.Any], NoValue]
    ):
        # type: (...) -> typing.Iterable[T]
        """
        Compare the left and right content of path in case both objects are list.
        """
        raise NotImplementedError()

    @abstractmethod
    def value_check(
        self,
        path,  # type: PathType
        left_value,  # type: typing.Any
        right_value,  # type: typing.Any
    ):
        # type: (...) -> typing.Iterable[T]
        """
        Compare the left and right content of path in case the objects have different types or are not dictionaries or lists.
        """
        raise NotImplementedError()

    def _is_recursive_call(self, path, left, right):
        # type: (PathType, typing.Any, typing.Any) -> bool
        """
        Determine if the current objects are already been traversed.

        Swagger specification could contain recursive definitions and references.
        Due to the fact that we fully dereference the specs then there will be no
        good wait to know if we've already visited the given objects other than
        using their ids.
        """
        cache_key = (id(left), id(right))

        if cache_key in self._inner_walk_calls and any(
            path[:len(known_path)] == known_path for known_path in self._inner_walk_calls[cache_key]
        ):
            # Deal with recursive objects
            return True

        self._inner_walk_calls[cache_key].append(path)
        return False

    def _inner_walk(self, path, left, right):
        # type: (PathType, typing.Any, typing.Any) -> typing.Iterable[T]
        """
        Fully traverse the left and right objects.

        The traversal will short-circuit in case:
         * a given path should not be traversed
         * the path has already been traversed (recursive definition)
        """
        if not self.should_path_be_walked_through(path) or self._is_recursive_call(path, left, right):
            return ()

        if isinstance(left, dict) and isinstance(right, dict):
            return chain(
                self.dict_check(path, left, right),
                (
                    value
                    for key in set(chain(iterkeys(left), iterkeys(right)))
                    for value in self._inner_walk(
                        path=tuple(chain(path, [key])),
                        left=left.get(key, NO_VALUE),
                        right=right.get(key, NO_VALUE),
                    )
                ),
            )
        elif isinstance(left, list) and isinstance(right, list):
            return chain(
                self.list_check(path, left, right),
                (
                    value
                    for index, (left_item, right_item) in enumerate(zip_longest(left, right, fillvalue=NO_VALUE))
                    for value in self._inner_walk(
                        path=tuple(chain(path, [index])),
                        left=left_item,
                        right=right_item,
                    )
                ),
            )
        else:
            return self.value_check(path, left, right)

    def walk(self):
        # type: () -> typing.Iterable[T]
        """
        Fully traverse the left and right objects.
        NOTE:   the traversing is internally cached such that all the subsequent calls
                to `walk()` are equivalent to an attribute access
        """
        if isinstance(self._walk_result, NoValue):  # pragma: no branch
            self._walk_result = list(self._inner_walk(
                path=tuple(),
                left=self.left,
                right=self.right,
            ))
        return self._walk_result


class SchemaWalker(Walker[T]):
    """
    Walker aware of how a Swagger schema looks like.

    The main difference between this walker and Walker is that this walker
    keeps in consideration some peculiarity of the swagger specs.

    The walker implementation should never worry about dereferencing as the traversing
    is performed on the fully flattened and dereferenced specs
    """

    def __init__(self, left_spec, right_spec, **kwargs):
        # type: (Spec, Spec, typing.Any) -> None
        super(SchemaWalker, self).__init__(
            left=left_spec.deref_flattened_spec,
            right=right_spec.deref_flattened_spec,
            left_spec=left_spec,
            right_spec=right_spec,
            **kwargs
        )

    def _is_path_a_parameter_list_location(self, path):
        # type: (PathType) -> bool
        """
        Check if the given path is compatible with a path that contains the list of
        parameters of an operation object.

        The possible locations are:
        1) /paths/<endpoint path>/parameters
            https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#path-item-object
        2) /paths/<endpoint path>/<http verb>/parameters
            https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#operation-object
        """
        if len(path) == 3 and path[0] == 'paths' and path[2] == 'parameters':
            return True
        elif len(path) == 4 and path[0] == 'paths' and path[2] in _HTTP_OPERATIONS and path[3] == 'parameters':
            return True

        return False

    def _get_original_parameter_path(self, path, parameters_index):
        # type: (PathType, typing.Mapping[typing.Text, int]) -> PathType
        try:
            # Ignoring type as path[-1] could be an integer which is not expected for parameters_index
            # it's not really a big deal as it should not happen and if this happens then KeyError will be thrown
            return tuple(chain(path[:-1], [parameters_index[path[-1]]]))  # type: ignore
        except KeyError:
            # This could happen only if the parameter was present only on the old specs
            # or if the path was actually not "modified" by the walker (NOTE: the later condition should not be possible)
            # but let's do it so mypy is happy too
            return path

    def fix_parameter_path(self, path, original_path, value):
        # type: (PathType, PathType, T) -> T
        """
        Fix an eventual path present on the value returned by the walker.

        The SwaggerAwareWalker modifies the indexing approach used for parameters due to the fact
        that parameters are defined as arrays and modifying their order would not change the semantic.
        """
        try:
            # ignoring type as we're exploiting duck typing and is not easy to validate the protocol at run time
            return value.fix_parameter_path(path=path, original_path=original_path)  # type: ignore
        except TypeError as type_error:
            warnings.warn(
                str('Unexpected {}.fix_parameter_path signature. {}'.format(
                    value.__class__.__name__,
                    type_error,
                )),
                category=RuntimeWarning,
            )
        except AttributeError:
            # Ignore such exception as it means that value does not implement fix_parameter_path method
            pass

        return value

    def _inner_walk(self, path, left, right):
        # type: (PathType, typing.Any, typing.Any) -> typing.Iterable[T]

        if self._is_path_a_parameter_list_location(path):
            left_parameters_map = {} if left is NO_VALUE else {parameter['name']: parameter for parameter in left}
            right_parameters_map = {} if right is NO_VALUE else {parameter['name']: parameter for parameter in right}
            parameters_index = {} if right is NO_VALUE else {parameter['name']: index for index, parameter in enumerate(right)}
            return (
                self.fix_parameter_path(
                    path=new_path,
                    original_path=self._get_original_parameter_path(new_path, parameters_index),
                    value=value,
                )
                for key in set(chain(iterkeys(left_parameters_map), iterkeys(right_parameters_map)))
                for new_path in (tuple(chain(path, [key])), )  # Small trick to allow variable definition in generator-comprehension
                for value in self._inner_walk(
                    path=tuple(chain(path, [key])),
                    left=left_parameters_map.get(key, NO_VALUE),
                    right=right_parameters_map.get(key, NO_VALUE),
                )
            )
        else:
            return super(SchemaWalker, self)._inner_walk(path=path, left=left, right=right)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import typing

from termcolor import colored

from swagger_spec_compatibility.cli.common import add_rules_arguments
from swagger_spec_compatibility.cli.common import CLIProtocol
from swagger_spec_compatibility.cli.common import rules


class _Namespace(CLIProtocol):
    blacklist_rules = None  # type: typing.Iterable[typing.Text]
    rules = None  # type: typing.Iterable[typing.Text]


def execute(cli_args):
    # type: (_Namespace) -> int
    print(
        '{title}\n{rules_detail}'.format(
            title=colored('Rules explanation', attrs=['bold']),
            rules_detail='\n'.join(
                rule.explain()
                for rule in sorted(rules(cli_args), key=lambda rule: rule.error_code)
            ),
        ),
    )

    return 0


def add_sub_parser(subparsers):
    # type: (argparse._SubParsersAction) -> argparse.ArgumentParser
    explain_parser = subparsers.add_parser('explain', help='explain selected rules')
    add_rules_arguments(explain_parser)
    return explain_parser
